__version__ = '0.9.0a8'

# Shortcuts to quick-use utils
from djem.utils.inspect import pp  # noqa: F401
from djem.utils.misc import UNDEFINED  # noqa: F401
from djem.utils.mon import M, Mon, mon  # noqa: F401
